"""
Goal: Store the rules by which we pick the elements of patient data before
    computing the hashed string.

Authors:
     Andrei Sura <sura.andrei@gmail.com>
"""

# TODO: research the claim from
# L. Sweeney. k-anonymity: a model for protecting privacy.
# International Journal on Uncertainty, Fuzziness and Knowledge-based Systems
#
# "...87% (216 million of 248 million) of the population in the
#   United States had reported characteristics that likely made them
#   unique based only on {5-digit ZIP, sex, date of birth}."

from onefl import utils  # noqa
from onefl.normalized_patient import NormalizedPatient  # noqa


# Last Name + First Name + DOB + Race
RULE_CODE_F_L_D_R = 'F_L_D_R'

# First Name + Last Name + DOB + Sex
RULE_CODE_F_L_D_S = 'F_L_D_S'


# In order to guarantee correctness we will allow the partners
# to add to the configuration only values from the map below.
# If we add new rules then we will ask the partners to download a new version
# of the client software.
AVAILABLE_RULES_MAP = {

    RULE_CODE_F_L_D_S: {
        'required_attr': ['pat_first_name', 'pat_last_name', 'pat_birth_date', 'pat_sex'],  # NOQA
        'pattern': '{0.pat_first_name}{0.pat_last_name}{0.pat_birth_date}{0.pat_sex}',  # NOQA
    },
    RULE_CODE_F_L_D_R: {
        'required_attr': ['pat_first_name', 'pat_last_name', 'pat_birth_date', 'pat_race'],  # NOQA
        'pattern': '{0.pat_last_name}{0.pat_first_name}{0.pat_birth_date}{0.pat_race}',  # NOQA
    },
}



"""
Goal: implement base class for tests using the database

@authors:
    Andrei Sura <sura.andrei@gmail.com>
"""

from unittest import TestCase  # noqa
from datetime import datetime
from sqlalchemy.orm.exc import NoResultFound  # noqa
from onefl.config import Config
from onefl.models import base
from onefl.utils import db
from sqlalchemy.orm.exc import MultipleResultsFound  # noqa
from onefl.models.partner_entity import PartnerEntity
from onefl.models.rule_entity import RuleEntity
from onefl.rules import RULE_CODE_F_L_D_R, RULE_CODE_F_L_D_S  # noqa

SETTINGS_FILE = 'config/test_settings_linker.py'


class BaseTestCase(TestCase):
    """
    Tests for database interaction extend this class.
    @see http://docs.sqlalchemy.org/en/latest/orm/session_transaction.html#session-external-transaction  # noqa
    """

    def __init__(self, *args, **kwargs):
        super(BaseTestCase, self).__init__(*args, **kwargs)

    def setUp(self):
        """ Set the test database engine/session """

        super(BaseTestCase, self).setUp()
        config = Config(root_path='.', defaults={})
        config.from_pyfile(SETTINGS_FILE)
        self.config = config
        self.engine = db.get_db_engine(config)
        self.session = db.get_db_session(self.engine,
                                         create_tables=True)
        self.create_rules()
        self.create_partners()

    def create_rules(self):
        """ Create rule rows """
        added_date = datetime.now()

        # with self.assertRaises(NoResultFound):
        #     self.session.query(RuleEntity).filter_by(id=1).one()

        rule_r = self.session.query(RuleEntity).filter_by(
            rule_code=RULE_CODE_F_L_D_R).one_or_none()
        rule_g = self.session.query(RuleEntity).filter_by(
            rule_code=RULE_CODE_F_L_D_S).one_or_none()

        if rule_r is None:
            RuleEntity.create(
                rule_code=RULE_CODE_F_L_D_R,
                rule_description='First Last DOB Race',
                rule_added_at=added_date)

        if rule_g is None:
            RuleEntity.create(
                rule_code=RULE_CODE_F_L_D_S,
                rule_description='First Last DOB Sex',
                rule_added_at=added_date)

        # self.assertEquals(2, rule.id)
        cache = RuleEntity.get_rules_cache(self.session)
        self.assertIsNotNone(cache)
        print(cache)

    def create_partners(self):
        """ Verify we can store PARTNER rows """
        # added_date = utils.get_db_friendly_date_time()
        added_date = datetime.now()

        partner_ufh = self.session.query(PartnerEntity).filter_by(
            partner_code='UFH').one_or_none()

        if partner_ufh is None:
            partner_ufh = PartnerEntity.create(
                partner_code="UFH",
                partner_description="University of Florida",
                partner_added_at=added_date)

            print("Saved fresh row: {}\n".format(partner_ufh))
        self.assertEquals("UFH", partner_ufh.partner_code)

        partner_flm = self.session.query(PartnerEntity).filter_by(
            partner_code='FLM').one_or_none()

        if partner_flm is None:
            partner_flm = PartnerEntity.create(
                partner_code="FLM",
                partner_description="Florida Medicaid",
                partner_added_at=added_date)
            self.assertEquals("FLM", partner_flm.partner_code)

        # verify an error is raised if we try to to find one
        with self.assertRaises(MultipleResultsFound):
            self.session.query(PartnerEntity).filter(
                PartnerEntity.partner_description.like(
                    '%Florida%')).one()

    def tearDown(self):
        """ Drop the tables created """
        super(BaseTestCase, self).tearDown()
        base.metadata.drop_all(self.engine)




from collections import OrderedDict

IN_DELIMITER = '\t'
OUT_DELIMITER = '\t'

IN_FILE = 'phi_hashes.csv'
OUT_FILE = 'links.csv'
SAVE_OUT_FILE = True

# How many lines to parse at once
LINES_PER_CHUNK = 2000

# List rules for computing hashes
ENABLED_RULES = OrderedDict(
    [('1', 'F_L_D_R'),
     ('2', 'F_L_D_S')])

# List column names for the input file
EXPECTED_COLS = ['PATID']
EXPECTED_COLS.extend(ENABLED_RULES.values())

# =============== Database Settings
DB_TYPE = 'TEST'



"""
Goal: store utility functions for user interaction

@authors:
    Andrei Sura <sura.andrei@gmail.com>
"""
import os
import sys


def check_file_exists(ask, file_name):
    """
    If ask = `True` prompt the user to continue or stop,
    otherwise pretend that the file exists.
    """
    exists = False

    if not os.path.isfile(file_name):
        if ask:
            opt = ask_yes_no("The file {} does not exist. "
                             "Continue anyway?".format(file_name))
            if not opt:
                sys.exit('Got it.')
    else:
        exists = True
    return exists


def ask_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return the answer
    as a boolean.

    :param question: the question displayed to the user
    :param default: the default answer if the user hits <Enter>

    """
    valid = {"y": True, "n": False}

    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()[0]

        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")



"""
Goal: Store functions used for converting PHI data into hashed strings

@authors:
    Andrei Sura <sura.andrei@gmail.com>
"""

import os
import pandas as pd
from onefl.rules import AVAILABLE_RULES_MAP as rulz
from onefl import utils
from onefl.exc import ConfigErr
from onefl.normalized_patient import NormalizedPatient  # noqa

pd.set_option('display.width', 1500)
VALID_RACE_VALS = ['', '01', '02', '03', '04', '05', '06', '07', 'NI', 'UN', 'OT']  # noqa
VALID_SEX_VALS = ['', 'A', 'F', 'M', 'NI', 'UN', 'OT']


class HashGenerator():
    log = None

    @classmethod
    def configure_logger(cls, logger):
        cls.log = logger

    @classmethod
    def _process_row_series(cls, ser, rule, pattern, required_attr, config):
        """
        :param config: dictionary with run-time parameters

        :rtype: string
        :return sha_string:
        """

        patient = NormalizedPatient(
            patid=ser['patid'],
            pat_first_name=ser['first'],
            pat_last_name=ser['last'],
            pat_birth_date=ser['dob'],
            pat_sex=ser['sex'],
            pat_race=ser['race']
        )
        # cls.log.info("Patient: {}".format(patient))

        if not patient.has_all_data(required_attr):
            cls.log.debug("Skip hashing patient [{}] due to missing data"
                          "for rule [{}]".format(patient.patid, rule))
            return ''

        raw = pattern.format(patient) + config['SALT']
        sha_string = utils.apply_sha256(raw)
        cls.log.debug("For patient [{}] (rule {}): {}, hash_string= {}".format(patient.patid, rule, raw, sha_string))  # noqa

        return sha_string

    @classmethod
    def _process_frame(cls, df_source, config):
        """
        Create a result frame

        Reminder:
            - apply() works on a row / column basis of a DataFrame
            - applymap() works element-wise on a DataFrame
            - map() works element-wise on a Series

        """
        df = pd.DataFrame()

        # keep the patid from the source
        df['PATID'] = df_source['patid']

        for i, rule in enumerate(rulz):
            cls.log.debug("Applying rule {}: {}".format(i, rule))
            rule_data = rulz.get(rule)
            pattern = rule_data['pattern']
            required_attr = rule_data['required_attr']

            df[rule] = df_source.apply(
                lambda x: cls._process_row_series(x, rule, pattern,
                                                  required_attr,
                                                  config), axis=1)
        cls.log.debug("Processed frame: \n{}".format(df))
        return df

    @classmethod
    def _validate_config(cls, config):
        """
        Helper method for preventing config errors
        """
        enabled_rules = config.get('ENABLED_RULES', None)

        if not enabled_rules:
            raise ConfigErr('Please verify that the config specifies'
                            ' the ENABLED_RULES parameter')

        for rule_code in enabled_rules:
            if rule_code not in rulz:
                raise ConfigErr('Invalid rule code: [{}]! '
                                'Available codes are: {}'
                                .format(rule_code, rulz.keys()))

    @classmethod
    def generate(cls, config, inputdir, outputdir):
        """
        Read the "phi.csv" file and generate "phi_hashes.csv"
        containing two (or more) sha256 strings for each line
        in the input file.

        This method is invoked from

        .. seealso::

            run_hasher.py

        :param inputdir: directory name for the source file
        :param outputdir: directory name for generated file

        :rtype: DataFrame
        :return the frame with hashes of the PHI data

        Columns:
            - patid
            - sha_rule_1 (first_last_dob_race)
            - sha_rule_2 (first_last_dob_sex)

        """
        cls._validate_config(config)
        EXPECTED_COLS = config['EXPECTED_COLS']

        cls.log.info("Using [{}] as source folder".format(inputdir))
        cls.log.info("Using [{}] as salt".format(config['SALT']))
        cls.log.info("Expecting input file to contain columns: {}"
                     .format(EXPECTED_COLS))
        cls.log.info("Using [{}] as destination folder".format(outputdir))

        in_file = os.path.join(inputdir, config['IN_FILE'])
        reader = None

        try:
            reader = pd.read_csv(in_file,
                                 sep=config['IN_DELIMITER'],
                                 dtype=object,
                                 skipinitialspace=True,
                                 skip_blank_lines=True,
                                 usecols=list(EXPECTED_COLS),
                                 chunksize=config['LINES_PER_CHUNK'],
                                 iterator=True)
            cls.log.info("Reading data from file: {} ({})"
                         .format(in_file, utils.get_file_size(in_file)))

        except ValueError as exc:
            cls.log.info("Please check if the actual column names"
                         " in [{}] match the expected column names"
                         " file: {}.".format(in_file,
                                             sorted(EXPECTED_COLS)))
            cls.log.error("Error: {}".format(exc))

        frames = []

        for df_source in reader:
            df_source.fillna('', inplace=True)

            for col in EXPECTED_COLS:
                if col not in sorted(df_source):
                    raise Exception("The input data frame does not have all "
                                    "expected columns: {}"
                                    .format(EXPECTED_COLS))

            # validate the values constrained to set
            invalid_race = df_source.loc[~df_source['race'].isin(VALID_RACE_VALS)]  # noqa
            invalid_sex = df_source.loc[~df_source['sex'].isin(VALID_SEX_VALS)]

            if len(invalid_race) > 0:
                cls.log.info("Please check race: {}".format(invalid_race))
                raise Exception("The input file contains invalid value for "
                                "`race` column. Please review the specs.")
            if len(invalid_sex) > 0:
                cls.log.warning("Please check sex: {}".format(invalid_sex))
                raise Exception("The input file contains invalid value for "
                                "`sex` column. Please review the specs.")

            df = cls._process_frame(df_source, config)
            frames.append(df)

        df = pd.concat(frames, ignore_index=True)

        # Concatenation can re-order columns so we need to enforce the order
        out_columns = ['PATID']
        out_columns.extend(config.get('ENABLED_RULES'))

        out_file = os.path.join(outputdir, config['OUT_FILE'])
        utils.frame_to_file(df[out_columns], out_file,
                            delimiter=config['OUT_DELIMITER'])

        cls.log.info("Wrote output file: {} ({} data rows, {})"
                     .format(out_file,
                             len(df),
                             utils.get_file_size(out_file)))
        return True



"""
ORM for "linkage" table

@authors:
    Andrei Sura <sura.andrei@gmail.com>
"""
import sqlalchemy as db
import binascii
from onefl import utils
from onefl.models.base import DeclarativeBase
from onefl.models.crud_mixin import CRUDMixin
from onefl.models.partner_entity import PartnerEntity
from onefl.models.rule_entity import RuleEntity  # noqa

"""
+------------------+---------------------+------+-----+---------+
| Field            | Type                | Null | Key | Default |
+------------------+---------------------+------+-----+---------+
| linkage_id       | bigint(20) unsigned | NO   | PRI | NULL    |
| partner_code     | varchar(3)          | NO   | MUL | NULL    |
| linkage_patid    | varchar(128)        | NO   | MUL | NULL    |
| linkage_flag     | int                 | NO   | MUL | NULL    |
| linkage_uuid     | binary(16)          | NO   | MUL | NULL    |
| linkage_hash     | binary(32)          | NO   | MUL | NULL    |
| linkage_added_at | datetime            | NO   | MUL | NULL    |
+------------------+---------------------+------+-----+---------+
"""

__all__ = ['LinkageEntity']


class LinkageEntity(CRUDMixin, DeclarativeBase):

    """ Maps the UUIDs to "hashed chunks" """
    __tablename__ = 'LINKAGE'

    id = db.Column('LINKAGE_ID', db.Integer, primary_key=True)
    partner_code = db.Column('PARTNER_CODE', db.Text(3),
                             db.ForeignKey('PARTNER.PARTNER_CODE'),
                             nullable=False)

    rule_id = db.Column('RULE_ID', db.Integer,
                        db.ForeignKey('RULE.RULE_ID'), nullable=False)

    # This column stores the original (unscrambled) patid
    linkage_patid = db.Column('LINKAGE_PATID', db.Text(128), nullable=False)
    linkage_flag = db.Column('LINKAGE_FLAG', db.Integer, nullable=False)

    # The generated ID that de-duplicates records
    linkage_uuid = db.Column('LINKAGE_UUID', db.Binary, nullable=False)

    # The hashed representation of the patient data
    linkage_hash = db.Column('LINKAGE_HASH', db.Binary, nullable=False)

    # timestamp
    linkage_added_at = db.Column('LINKAGE_ADDED_AT', db.DateTime,
                                 nullable=False)

    # @OneToOne
    partner = db.orm.relationship(PartnerEntity, uselist=False, lazy='joined')
    rule = db.orm.relationship(RuleEntity, uselist=False, lazy='joined')

    @staticmethod
    def short_hash(val):
        return val[:8]

    def friendly_uuid(self):
        return utils.hexlify(self.linkage_uuid)

    def friendly_hash(self):
        return utils.hexlify(self.linkage_hash)

    @staticmethod
    def init_hash_uuid_lut(session, hashes):
        """
        From the list [x, y, z] of hashes return
        a dictionary which tells if a chunk was `linked` or not:
            {x: LinkageEntity, y: LinkageEntity, z: None}
        """
        # Note: unhexlify is necessary since the database stores
        # binary representations of the hashes
        bin_hashes = [binascii.unhexlify(ahash.encode('utf-8'))
                      for ahash in hashes]
        links = session.query(LinkageEntity).filter(
            LinkageEntity.linkage_hash.in_(bin_hashes)).all()
        links_cache = {link.friendly_hash(): link for link in links}

        lut = {}

        # Extra loop so we can provide an entry for hashes not found in the db
        for ahash in hashes:
            lut[ahash] = links_cache.get(ahash, None)

        return lut

    def __repr__(self):
        """ Return a friendly object representation """

        return "<LinkageEntity (linkage_id: {0.id}, "\
            "partner_code: {0.partner_code}, " \
            "rule_id: {0.rule_id}, "\
            "linkage_patid: {0.linkage_patid}, "\
            "linkage_flag: {0.linkage_flag}, "\
            "linkage_uuid: {1}, "\
            "linkage_hash: {2}, "\
            "linkage_addded_at: {0.linkage_added_at})>".format(
                self,
                binascii.hexlify(self.linkage_uuid),
                binascii.hexlify(self.linkage_hash)
            )



#!/usr/bin/env python
"""
Goal: Implement the entry point for the hash generator

@authors:
    Andrei Sura <sura.andrei@gmail.com>
"""

import time
import argparse
from datetime import timedelta
from onefl import logutils
from onefl.config import Config
from onefl.hash_generator import HashGenerator
from onefl.normalized_patient import NormalizedPatient
from onefl.version import __version__

logger = logutils.get_a_logger(__file__)

# The config file is looked up relative to this "root" folder
ROOT_PATH = '.'
DEFAULT_SETTINGS_FILE = 'config/settings_hasher.py'


def main():
    """
    Configure the logger object and read the command line arguments
    for invoking the generator.

    .. seealso::

        :meth:`HashGenerator.generate`

    """
    HashGenerator.configure_logger(logger)
    NormalizedPatient.configure_logger(logger)

    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version",
                        default=False,
                        action='store_true',
                        help="Show the version number")

    parser.add_argument("-c", "--config",
                        default=DEFAULT_SETTINGS_FILE,
                        help="Application config file")
    parser.add_argument(
        '-i', '--inputdir',
        # required=True,
        default='.',
        help='input directory name')
    parser.add_argument(
        '-o', '--outputdir',
        # required=True,
        default='.',
        help='output directory name')

    args = parser.parse_args()

    if args.version:
        import sys
        print("deduper, version {}".format(__version__))
        sys.exit()

    config = Config(root_path=ROOT_PATH, defaults={})
    config.from_pyfile(args.config)
    start = time.monotonic()
    success = HashGenerator.generate(config, args.inputdir, args.outputdir)
    end = time.monotonic()
    elapsed = (end - start)

    if success:
        logger.info("Done. Process duration: {}"
                    .format(str(timedelta(seconds=elapsed))))
    else:
        logger.error("Failed!")

if __name__ == '__main__':
    main()



"""
Goal: implement tests for `HashGenerator` class

@authors:
    Andrei Sura <sura.andrei@gmail.com>
"""
import os
import unittest
import pandas.util.testing as tm
# from mock import patch
# from io import StringIO
from onefl.config import Config
from onefl import logutils
from onefl import utils
from onefl.utils.ui import check_file_exists
from onefl.hash_generator import HashGenerator
from onefl.normalized_patient import NormalizedPatient

# patch it
# sys.exit = lambda *x: None

DELIMITER = '\t'
logger = logutils.get_a_logger(__file__)

SETTINGS_FILE = 'config/test_settings_hasher.py'


class TestHashGenerator(unittest.TestCase):

    def setUp(self):
        super(TestHashGenerator, self).setUp()
        HashGenerator.configure_logger(logger)
        NormalizedPatient.configure_logger(logger)

    def tearDown(self):
        pass

    # @unittest.skip
    def test_hasher(self):
        """ Verify that we produce hashes properly """
        inputdir = 'tests/data_in'
        outputdir = 'tests/data_out'
        config = Config(root_path='.', defaults={})
        config.from_pyfile(SETTINGS_FILE)
        result = HashGenerator.generate(config, inputdir, outputdir)
        self.assertTrue(result)

        # Check if the reference file exists
        file_name = os.path.join('tests/data_in', 'phi_hashes.csv')
        exists = check_file_exists(ask=False, file_name=file_name)
        self.assertTrue(exists)

        # read the reference frame
        df_expected = utils.frame_from_file(file_name, delimiter=DELIMITER)

#         file = StringIO(u"""
# patid| sha_rule_1| sha_rule_2
# 01 | 1 | 1
# 02 | 1 | 1
# 03 | 2 | 1
# 04 | 2 | 1
# """)
#         df_expected = pd.read_csv(file, sep='|',
#                                   dtype=object,
#                                   skipinitialspace=True,
#                                   skip_blank_lines=True)

        # read the frame produced by calling `generate()`
        df_actual = utils.frame_from_file(
            os.path.join('tests/data_out', 'phi_hashes.csv'),
            delimiter=DELIMITER)

        tm.assert_frame_equal(df_expected, df_actual)


if __name__ == '__main__':
    unittest.main()



"""
Goal: simplify the code when interacting with entities
"""
from onefl.models.base import get_session


class CRUDMixin():
    """ Helper class sqlalchemy entities """
    __table_args__ = {'extend_existing': True}

    # id = db.Column(db.Integer, primary_key=True)

    @classmethod
    def get_by_id(cls, id):
        if any(
            (isinstance(id, str) and id.isdigit(),
             isinstance(id, (int, float))),):
            session = get_session()
            # return cls.query.get(int(id))
            # return session.query(cls).filter_by(id=int(id)).one()
            return session.query(cls).get(int(id))
        return None

    @classmethod
    def create(cls, **kwargs):
        """ Helper for add() + commit() """
        instance = cls(**kwargs)
        return instance.save()

    def update(self, commit=True, **kwargs):
        """ update object instance """
        for attr, value in kwargs.items():
            setattr(self, attr, value)
        return self.save() if commit else self

    def save(self, commit=True):
        """ save object instance """
        session = get_session()
        session.add(self)

        if commit:
            session.commit()
        return self

    def delete(self, commit=True):
        """ delete object instance """
        session = get_session()
        session.delete(self)
        return commit and session.commit()



"""
Goal: store utility functions for database interaction

@authors:
    Andrei Sura <sura.andrei@gmail.com>
"""
# import sys
import sqlalchemy as db
from onefl.models import base
from onefl import logutils
log = logutils.get_a_logger(__file__)


DB_TYPE_MSSQL = 'MSSQL'
DB_TYPE_MYSQL = 'MYSQL'
DB_TYPE_TEST = 'TEST'

# try:
#     from config.settings import DB_USER, DB_PASS, DB_NAME, DB_HOST, DB_PORT
# except:
#     sys.exit("Please create the config/settings.py file. \n"
#              "  $ cp config/settings.py.example config/settings.py \n")

DB_POOL = None


def get_db_url_sqlserver(config):
    """
    Helper function for creating the "pyodbc" connection string.

    @see http://docs.sqlalchemy.org/en/latest/dialects/mssql.html
    @see https://code.google.com/p/pyodbc/wiki/ConnectionStrings
    """
    from urllib import parse
    params = parse.quote(
        "Driver={{FreeTDS}};Server={};Port={};"
        "Database={};UID={};PWD={};"
        .format(config['DB_HOST'], config['DB_PORT'], config['DB_NAME'],
                config['DB_USER'], config['DB_PASS']))
    return 'mssql+pyodbc:///?odbc_connect={}'.format(params)


def get_db_url_mysql(config):
    """
    Format the configuration parameters to build the connection string
    """
    if 'DB_URL_TESTING' in config:
        return config['DB_URL_TESTING']

    return 'mysql+mysqlconnector://{}:{}@{}/{}' \
           .format(config['DB_USER'],
                   config['DB_PASS'],
                   config['DB_HOST'],
                   config['DB_NAME'])


def get_db_url(config):
    """
    Pick a database url based on DATABASE_TYPE
    """
    db_type = config.get('DB_TYPE').upper()

    if DB_TYPE_MSSQL == db_type:
        url = get_db_url_sqlserver(config)
    elif DB_TYPE_MYSQL == db_type:
        url = get_db_url_mysql(config)
    elif DB_TYPE_TEST == db_type:
        # to use memory: "sqlite:///:memory:"
        url = 'sqlite:///sqlite.db'
    else:
        raise Exception("Unexpected config value for 'DB_TYPE': {}"
                        .format(db_type))
    return url


def get_db_engine(config):
    """
    @see http://docs.sqlalchemy.org/en/latest/core/connections.html
    """
    url = get_db_url(config)

    try:
        engine = db.create_engine(url,
                                  pool_size=10,
                                  max_overflow=5,
                                  pool_recycle=3600,
                                  echo=False)
    except TypeError as exc:
        log.warning("Got exc from db.create_engine(): {}".format(exc))
        engine = db.create_engine(url, echo=False)

    return engine


def get_db_session(engine, create_tables=False):
    """
    Connect to the database and return a Session object

    :param create_tables: boolean used to request table creation
    """
    log.debug("Call get_db_session()")
    base.init(engine)
    session = base.DBSession()

    if create_tables:
        base.metadata.create_all(engine)

    return session

# def get_db_url(db_host, db_port, db_name, db_user, db_pass):
#     """
#     Helper function for creating the "pyodbc" connection string.
#
#     @see http://docs.sqlalchemy.org/en/latest/dialects/mssql.html
#     @see https://code.google.com/p/pyodbc/wiki/ConnectionStrings
#     """
#     params = parse.quote(
#         "Driver={{FreeTDS}};Server={};Port={};"
#         "Database={};UID={};PWD={};"
#         .format(db_host, db_port, db_name, db_user, db_pass))
#     return 'mssql+pyodbc:///?odbc_connect={}'.format(params)
#
#
# def get_db_engine():
#     """
#     Create the database engine connection.
#     @see http://docs.sqlalchemy.org/en/latest/core/engines.html
#
#     :return: Dialect object which can either be used directly
#             to interact with the database, or can be passed to
#             a Session object to work with the ORM.
#     """
#     global DB_POOL
#
#     if DB_POOL is None:
#         url = get_db_url(db_host=DB_HOST, db_port=DB_PORT, db_name=DB_NAME,
#                          db_user=DB_USER, db_pass=DB_PASS)
#         DB_POOL = db.create_engine(url,
#                                    pool_size=10,
#                                    max_overflow=5,
#                                    pool_recycle=3600)
#
#     try:
#         DB_POOL.execute("USE {db}".format(db=DB_NAME))
#     except db.exc.OperationalError:
#         log.error('Database {db} does not exist.'.format(db=DB_NAME))
#
#     return DB_POOL




import unittest
# from mock import patch
from onefl import utils


class TestUtils(unittest.TestCase):

    def test_prepare_for_hashing(self):
        """ Verify that punctuation characters are removed """
        subjects = {
            '1': {'in': 'AbC xyZă', 'out': 'abcxyză'},
            '2': {'in': 'A&B,C.D:E;F-G}H{I@#J!', 'out': 'abcdefghij'},
            '3': {'in': 'ABC!"#$%&\'()*+,-./:;=>?@[\\]^_`{|}~ \t\n\r\x0b\x0c', 'out': 'abc'},  # NOQA
        }

        for case, data in subjects.items():
            self.assertEqual(data.get('out'),
                             utils.prepare_for_hashing(data.get('in')))



"""
Goal: implement tests for `LinkGenerator` class

@authors:
    Andrei Sura <sura.andrei@gmail.com>
"""
# flake8: noqa
import os
import unittest
import pandas.util.testing as tm  # noqa
from base_test import BaseTestCase
from onefl.config import Config
from onefl import logutils
from onefl import utils
from onefl.utils.ui import check_file_exists
from onefl.link_generator import LinkGenerator

DELIMITER = '\t'
logger = logutils.get_a_logger(__file__)

SETTINGS_FILE = 'config/test_settings_linker.py'


# class TestLinkGenerator(unittest.TestCase):
class TestLinkGenerator(BaseTestCase):

    def setUp(self):
        super(TestLinkGenerator, self).setUp()
        LinkGenerator.configure_logger(logger)

    def tearDown(self):
        pass

    def test_linker(self):
        """ Verify that we produce hashes properly """
        inputdir = 'tests/data_in'
        outputdir = 'tests/data_out'
        config = Config(root_path='.', defaults={})
        config.from_pyfile(SETTINGS_FILE)
        result = LinkGenerator.generate(config, inputdir, outputdir,
                                        partner='UFH')
        self.assertTrue(result)

        # Check if the reference file exists
        file_name = os.path.join(inputdir, 'links.csv')
        exists = check_file_exists(ask=False, file_name=file_name)
        self.assertTrue(exists)

        # read the reference frame
        df_expected = utils.frame_from_file(file_name, delimiter=DELIMITER)

        # read the frame produced by calling `generate()`
        file_name_actual = os.path.join(outputdir, 'links.csv')
        df_actual = utils.frame_from_file(file_name_actual,
                                          delimiter=DELIMITER)

        # tm.assert_frame_equal(df_expected, df_actual)
        uuids_actual = set()
        uuids_expected = set()
        uuids_actual.update(df_actual['UUID'].tolist())
        uuids_expected.update(df_expected['UUID'].tolist())

        # compare the numbers of distinct UUIDs
        self.assertEqual(len(uuids_actual), len(uuids_expected))

        # when we run the second time there are rows in the database
        # so we should get no new UUIDs...
        result = LinkGenerator.generate(config, inputdir, outputdir,
                                        partner='FLM')

        df_actual_2 = utils.frame_from_file(file_name_actual,
                                            delimiter=DELIMITER)
        # print(df_actual_2)




if __name__ == '__main__':
    unittest.main()



"""
Goal: implement the base class for models
"""
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.ext import declarative

# http://docs.sqlalchemy.org/en/latest/orm/session_basics.html
maker = sessionmaker(autoflush=True,
                     autocommit=False,
                     expire_on_commit=True)
DBSession = scoped_session(maker)
session = None


def init(engine):
    """
    Call this method before using any of the ORM classes.

    :seealso: :meth:get_db_session()
    """
    DBSession.configure(bind=engine)


def get_session():
    global session
    if session is None:
        session = DBSession()
    return session

DeclarativeBase = declarative.declarative_base()
metadata = DeclarativeBase.metadata



"""
Goal: provide access to the version information
"""

try:
    import pkg_resources
    __version__ = pkg_resources.require("deduper")[0].version
except Exception:
    from setuptools_scm import get_version
    __version__ = get_version()


if __name__ == "__main__":
    print(__version__)



"""
Goal: store utility functions not specific to a module

@authors:
    Andrei Sura <sura.andrei@gmail.com>
"""
import pandas as pd
import os
import sys
import unicodedata
# import hashlib
import uuid
import binascii
from hashlib import sha256

from csv import QUOTE_NONE, QUOTE_ALL  # noqa
# from urllib import parse
# from datetime import timedelta
# from datetime import datetime
# from base64 import b64decode, b64encode
from datetime import datetime
from datetime import date

from onefl import logutils
log = logutils.get_a_logger(__file__)


FORMAT_DATABASE_DATE = "%Y-%m-%d"
# import locale
# locale.setlocale(locale.LC_ALL, 'en_US')
ESCAPECHAR = '\\'


# table of punctuation characters + space
CHARS_TO_DELETE = dict.fromkeys(
    i for i in range(sys.maxunicode)
    if unicodedata.category(chr(i)).startswith('P') or
    not chr(i).isalnum())


def prepare_for_hashing(text):
    """
    Given a string with punctuation characters
    """
    if not text:
        return ''
    return text.translate(CHARS_TO_DELETE).lower()


def hexlify(val):
    """
    This function is used to display binary data in a friendly format.

        .. seealso::
            :meth:`LinkageEntity:friendly_uuid`

    Note:
        - Without the decode() the builtin `hexlify` return the bytes for
            hexadecimal representation of the binary data.
        - The returned string is twice as long as the length of data.

    :param val: binary
    :rtype: string
    """
    return binascii.hexlify(val).decode()


def get_uuid_bin(uuid_text=None):
    """
    Note: the returned value needs to be hexlified to be human readable
    """
    if not uuid_text:
        uuid_text = uuid.uuid1()
    return binascii.unhexlify(str(uuid_text).replace('-', '').lower().encode())


def apply_sha256(val):
    """ Compute sha256 sum
    :param val: the input string
    :rtype string: the sha256 hexdigest
    """
    m = sha256()
    m.update(val.encode('utf-8'))
    return m.hexdigest()


def format_date_as_string(val, fmt='%m-%d-%Y'):
    """
    :rtype str:
    :return the input value formatted as '%Y-%m-%d'

    :param val: datetime or string
    :param fmt: the input format for the date
    """
    if isinstance(val, date):
        return val.strftime(fmt)

    da = format_date(val, fmt)

    if not da:
        return ''

    return da.strftime(FORMAT_DATABASE_DATE)


def format_date(val, fmt='%m-%d-%Y'):
    """
    Transform the input string to a datetime object

    :param val: the input string for date
    :param fmt: the input format for the date
    """
    date_obj = None

    try:
        date_obj = datetime.strptime(val, fmt)
    except Exception as exc:
        log.warning("Problem formatting date: {} - {} due: {}"
                    .format(val, fmt, exc))

    return date_obj


def get_db_friendly_date_time():
    """
    :rtype: string
    :return current time in format: "2014-06-24 01:23:24"
    """
    return datetime.now().strftime('%Y-%m-%d %H:%M:%S')


def get_file_size(file_name):
    """
    :rtype numeric: the number of bytes in a file
    """
    bytes = os.path.getsize(file_name)
    return humanize_bytes(bytes)


def humanize_bytes(bytes, precision=1):
    """Return a humanized string representation of a number of bytes."""
    # This was stollen from http://code.activestate.com/recipes/577081/
    abbrevs = (
        (1 << 50, 'PB'),
        (1 << 40, 'TB'),
        (1 << 30, 'GB'),
        (1 << 20, 'MB'),
        (1 << 10, 'kB'),
        (1, 'bytes')
    )
    if bytes == 1:
        return '1 byte'

    for factor, suffix in abbrevs:
        if bytes >= factor:
            break
    return '%.*f %s' % (precision, bytes / factor, suffix)


def frame_from_file(file_name,
                    delimiter,
                    dtype=object,
                    quoting=QUOTE_NONE,
                    engine=None,
                    converters=None,
                    usecols=None,
                    index_col=None,
                    nrows=None,
                    error_bad_lines=False):
    """
    Load data frame from file.
    @see frame_to_file()

    API note: "error_bad_lines": boolean, default True
    Lines with too many fields (e.g. a csv line with too many commas)
    will by default cause an exception to be raised, and no DataFrame
    will be returned. If False, then these 'bad lines' will dropped
    from the DataFrame that is returned. (Only valid with C parser)

    """
    df = pd.read_csv(file_name,
                     sep=delimiter,
                     escapechar=ESCAPECHAR,
                     quoting=quoting,
                     dtype=dtype,
                     engine=engine,
                     converters=converters,
                     usecols=usecols,
                     index_col=index_col,
                     nrows=nrows,
                     error_bad_lines=error_bad_lines,
                     skipinitialspace=True,
                     skip_blank_lines=True,
                     low_memory=True,
                     )

    df.fillna('', inplace=True)
    df = df.applymap(str.strip)
    return df


def frame_to_file(df, file_name, delimiter="|"):
    """
    Store the dataframe to a file
    """
    if "|" == delimiter:
        log.warning("Writing frame {} using the default (|) delimiter"
                    .format(file_name))

    try:
        df.to_csv(file_name,
                  sep=delimiter,
                  escapechar=ESCAPECHAR,
                  index=False,
                  quoting=QUOTE_NONE,
                  encoding='utf-8')
    except Exception as exc:
        log.error("Unable to write frame due: {}".format(exc))

    return True



"""
Goal: store application settings
"""

# TODO: check if we need to allow flexible column naming
EXPECTED_COLS = ['patid', 'first', 'last', 'dob', 'race', 'sex']

IN_DELIMITER = '\t'
OUT_DELIMITER = '\t'

IN_FILE = 'phi.csv'
OUT_FILE = 'phi_hashes.csv'

# How many lines to parse at once
LINES_PER_CHUNK = 2000

# TODO: check if we need a secret salt that only partners know
# This would mean that we need to let partners
# exchange this string without ever leaking it to us
SALT = ''

# List rules for computing hashes
ENABLED_RULES = ['F_L_D_R', 'F_L_D_S']



"""
Goal: store shortcuts to common tasks

@authors:
    Andrei Sura <sura.andrei@gmail.com>

"""

# import sys
from invoke import task


@task
def list(ctx):
    """ Show available tasks """
    ctx.run('inv -l')


@task
def prep_develop(ctx):
    """ Install the requirements """
    ctx.run('pip install -U -r requirements.txt')
    print("==> Pip packages installed:")
    ctx.run('pip freeze')


@task(aliases=['hash'])
def hasher(ctx):
    """ Generate hashes from PHI """
    inputfolder = 'data'
    outputfolder = 'data'

    ctx.run('PYTHONPATH=. python run/hasher.py -i {} -o {} '
            .format(inputfolder, outputfolder))


@task(aliases=['link'])
def linker(ctx):
    """ Generate OneFlorida Ids from hashes """
    inputfolder = 'data'
    outputfolder = 'data'

    ctx.run('PYTHONPATH=. python run/linker.py -i {} -o {} '
            .format(inputfolder, outputfolder))


@task
def lint(ctx):
    """ Show the lint score """
    ctx.run("which pylint || pip install pylint")
    ctx.run("pylint -f parseable onefl")


@task
def test(ctx):
    """ Run tests
    -s: shows details
    """
    ctx.run('PYTHONPATH="." py.test -v -s --tb=short tests/ --color=yes')


@task(aliases=['cov'])
def coverage(ctx):
    """ Create coverage report """
    ctx.run('PYTHONPATH="." py.test -v -s --tb=short tests/ --color=yes'
            ' --cov onefl ')  # --cov-config tests/.coveragerc')


@task(aliases=['cov_html'])
def coverage_html(ctx):
    """ Create coverage report and open it in the browser"""
    ctx.run('PYTHONPATH="." py.test --tb=short -s --cov onefl '
            ' --cov-report term-missing --cov-report html tests/')
    ctx.run('open htmlcov/index.html')


@task
def devel_install(ctx):
    ctx.run('python setup.py develop')


@task
def devel_uninstall(ctx):
    ctx.run('python setup.py develop -u')


@task
def sdist(ctx):
    """ Create a `source` distribution """
    ctx.run("python setup.py sdist")


@task
def pypi_check_config(ctx):
    ctx.run("echo 'Check the presence of the ~/.pypirc config file'")
    ctx.run("test -f ~/.pypirc || echo 'Please create the ~/.pypirc file. "
            "Here is a template: \n'", echo=False)
    ctx.run("(test -f ~/.pypirc && echo "")|| (cat config/pypirc && exit 1)",
            echo=False)


@task(pre=[pypi_check_config])
def pypi_register(ctx):
    """ Use the ~/.pypirc config to register the package """
    ctx.run("python setup.py register -r deduper")


@task(pre=[pypi_check_config])
def pypi_upload(ctx):
    """ Use the ~/.pypirc config to upload the package """
    ctx.run("which twine || pip install twine")
    ctx.run("python setup.py sdist --formats=zip bdist_wheel")
    ctx.run("twine upload dist/* -r deduper")
    print("Done. To test please run: "
          "python -m virtualenv venv "
          " && source ./venv/bin/activate "
          " && pip install -U deduper && hasher -v")


@task
def pypi_wheel(ctx):
    """
    Download all `wheel` packages
    If the target machine has no access to internet we can build, ship, install
        pip wheel -r requirements-to-freeze.txt -w BUILT_WHEELS
        scp -r BUILT_WHEELS production_server:WHEE
        pip install -r requirements-to-freeze.txt --no-index --find-links WHEE
    """
    ctx.run('pip wheel deduper -r requirements-to-freeze.txt -w BUILT_WHEELS')


@task
def clean(ctx):
    """ Remove all generated files """
    ctx.run('find . -type f -name "*.pyc" -print | xargs rm -f')
    ctx.run('rm -rf htmlcov/ .coverage pylint.out')
    ctx.run('rm -rf .tox/* .ropeproject/')
    ctx.run('rm -rf ./dist ./build ./.eggs ./*.egg-info ./BUILT_WHEELS')
    # ctx.run('rm -f db.sqlite')


@task
def clean_log(ctx):
    """ Remove log file """
    ctx.run('rm -f logs/deduper.log')



"""
Goal: Lookup hashed strings in the database
    and generate a new uuid or return an existing one.

@authors:
    Andrei Sura <sura.andrei@gmail.com>
"""
import collections  # noqa
import os
import pandas as pd
import pprint  # noqa

from binascii import unhexlify
from datetime import datetime  # noqa

from onefl import utils
from onefl.utils import db
from onefl.exc import ConfigErr
from onefl.models.linkage_entity import LinkageEntity
from onefl.models.rule_entity import RuleEntity  # noqa
# from onefl.rules import AVAILABLE_RULES_MAP as rulz
from onefl.rules import RULE_CODE_F_L_D_R, RULE_CODE_F_L_D_S  # noqa

pd.set_option('display.width', 1500)

FLAG_HASH_NOT_FOUND = 0  # 'hash not found'
FLAG_HASH_FOUND = 1  # 'hash found'
FLAG_HASH_FOUND_BUT_IGNORED = 2  # 'hash found but ignored'


class LinkGenerator():
    log = None

    @classmethod
    def configure_logger(cls, logger):
        cls.log = logger

    @classmethod
    def _populate_hash_uuid_lut(cls, config, session, df_source):
        """
        Create a hash->UUID lookup-table for a set of patients
        we are parsing as a group.
        Note: the intent here is to reduce the number of database lookups.

       .. seealso::
           :meth:`LinkageEntity:init_hash_uuid_lut`
        """
        rules = config['ENABLED_RULES'].values()
        hashes = set()

        for rule in rules:
            hashes.update(df_source[rule].tolist())

        cls.log.info("Searching UUID's for {} distinct hashes."
                     .format(len(hashes)))
        hash_uuid_lut = LinkageEntity.init_hash_uuid_lut(session,
                                                         list(hashes))
        # distinct_uuids = {link.friendly_uuid() for link in
        #                   hash_uuid_lut.values() if link is not None}

        # cls.log.info("Found {} distinct UUID's from {} hashes."
        #              .format(len(distinct_uuids), len(hashes)))

        return hash_uuid_lut

    @classmethod
    def _process_patient_row(cls, patid, pat_hashes, hash_uuid_lut,
                             rules_cache, config, session,
                             partner_code):
        """
        :return OrderedDict: with the newly created linkage entities
        """
        links = {}
        to_investigate = {}

        # keep the hashes in the order defined by the rules
        # pat_hashes = sorted({row[rule] for rule in rules if row[rule] != ''})
        # pat_hashes = {rule: row[rule] for rule in rules if row[rule] != ''}

        if len(pat_hashes) == 1:
            # only one hash was received
            rule_code, ahash = pat_hashes.popitem()
            existing_link = hash_uuid_lut.get(ahash)
            binary_hash = unhexlify(ahash.encode('utf-8'))

            if existing_link is None:
                # create new UUID
                binary_uuid = utils.get_uuid_bin()
                flag = FLAG_HASH_NOT_FOUND
            else:
                # reuse the existing UUID
                binary_uuid = existing_link.linkage_uuid
                flag = FLAG_HASH_FOUND

            new_link = LinkageEntity.create(
                partner_code=partner_code,
                rule_id=rules_cache.get(rule_code),  # we need the rule_id here
                linkage_patid=patid,
                linkage_flag=flag,
                linkage_uuid=binary_uuid,
                linkage_hash=binary_hash,
                linkage_added_at=datetime.now())

            links = {ahash: new_link}

        elif len(pat_hashes) == 2:
            links, to_investigate = cls._process_two_hashes(
                patid, pat_hashes, hash_uuid_lut,
                rules_cache, config, session,
                partner_code)
        return links, to_investigate

    @classmethod
    def _process_two_hashes(cls, patid, pat_hashes, hash_uuid_lut,
                            rules_cache, config, session,
                            partner_code):
        links = {}
        added_date = datetime.now()

        # TODO: This is ugly but we can make (if needed)
        # the logic work for "n" rules
        rule_code_1, ahash_1 = pat_hashes.popitem()
        rule_code_2, ahash_2 = pat_hashes.popitem()
        existing_link_1 = hash_uuid_lut.get(ahash_1)
        existing_link_2 = hash_uuid_lut.get(ahash_2)

        if existing_link_1 is None:
            # respect the main rule - create two links with a `fresh` UUID
            binary_uuid = utils.get_uuid_bin()
            flag_1 = FLAG_HASH_NOT_FOUND

            # Flag 2 is set depending on presence of the link in the database
            if existing_link_2 is None:
                flag_2 = FLAG_HASH_NOT_FOUND
            else:
                flag_2 = FLAG_HASH_FOUND_BUT_IGNORED

            new_link_1 = LinkageEntity.create(
                partner_code=partner_code,
                rule_id=rules_cache.get(rule_code_1),
                linkage_patid=patid,
                linkage_flag=flag_1,
                linkage_uuid=binary_uuid,
                linkage_hash=unhexlify(ahash_1.encode('utf-8')),
                linkage_added_at=added_date)

            new_link_2 = LinkageEntity.create(
                partner_code=partner_code,
                rule_id=rules_cache.get(rule_code_2),
                linkage_patid=patid,
                linkage_flag=flag_2,
                linkage_uuid=binary_uuid,
                linkage_hash=unhexlify(ahash_2.encode('utf-8')),
                linkage_added_at=added_date)

        if existing_link_1 is not None:
            # respect the main rule - create two links with an `existing` UUID
            flag_1 = FLAG_HASH_FOUND

            if existing_link_2 is None:
                flag_2 = FLAG_HASH_NOT_FOUND
            else:
                if (existing_link_1.linkage_uuid ==
                        existing_link_2.linkage_uuid):
                    # consensus hence no problem
                    flag_2 = FLAG_HASH_FOUND
                else:
                    # the UUID's do not match - we need to investigate
                    to_investigate = {
                        ahash_1: existing_link_1.friendly_uuid(),
                        ahash_2: existing_link_2.friendly_uuid()
                    }
                    cls.log.warning("Hashes of the patid: {} are linked "
                                    "to two distinct UUIDs: {}"
                                    .format(patid, to_investigate))
                    return links, to_investigate

            # reuse the existing UUID
            binary_uuid = existing_link_1.linkage_uuid

            new_link_1 = LinkageEntity.create(
                partner_code=partner_code,
                rule_id=rules_cache.get(rule_code_1),
                linkage_patid=patid,
                linkage_flag=flag_1,
                linkage_uuid=binary_uuid,
                linkage_hash=unhexlify(ahash_1.encode('utf-8')),
                linkage_added_at=added_date)

            new_link_2 = LinkageEntity.create(
                partner_code=partner_code,
                rule_id=rules_cache.get(rule_code_2),
                linkage_patid=patid,
                linkage_flag=flag_2,
                linkage_uuid=binary_uuid,
                linkage_hash=unhexlify(ahash_2.encode('utf-8')),
                linkage_added_at=added_date)

        links[ahash_1] = new_link_1
        links[ahash_2] = new_link_2

        return links, {}

    @classmethod
    def _process_frame(cls, config, session, df_source, partner_code):
        """
        """
        # Init an empty frame and copy the patid from the source
        df = pd.DataFrame()
        df['PATID'] = df_source['PATID']
        hash_uuid_lut = cls._populate_hash_uuid_lut(config, session, df_source)
        mapped_hashes = {ahash: link.friendly_uuid()
                         for ahash, link in hash_uuid_lut.items()
                         if link is not None}
        rules_cache = RuleEntity.get_rules_cache(session)
        cls.log.debug("Found {} linked hashes in db: {}"
                      .format(len(mapped_hashes), mapped_hashes))

        # the rules are ordered by their importance
        rules = config['ENABLED_RULES'].values()
        patients_with_no_hashes = []

        for index, row in df_source.iterrows():
            patid = row['PATID']
            pat_hashes = {rule: row[rule] for rule in rules if row[rule] != ''}
            cls.log.debug("Parsing row for patient {} with {} hashes"
                          .format(patid, len(pat_hashes)))
            links, to_investigate = cls._process_patient_row(
                patid, pat_hashes.copy(), hash_uuid_lut,
                rules_cache, config, session,
                partner_code)

            i = 0
            for ahash, link in links.items():
                i += 1
                # print("Hash: {} link: {}".format(ahash, link))
                df.loc[df['PATID'] == patid, 'UUID'] = link.friendly_uuid()
                df.loc[df['PATID'] == patid, "hash_{}".format(i)] = ahash

            if len(pat_hashes) < 1:
                patients_with_no_hashes.append(patid)
                cls.log.warning("Patient [{}] has no hashes. "
                                "No UUID was created!".format(patid))
            else:
                cls.log.info("Created {} links for patid: {}"
                             .format(len(links), patid))
        cls.log.info("{} out of {} patients did not have any hashes: {}"
                     .format(len(patients_with_no_hashes), len(df),
                             patients_with_no_hashes))
        return df

    @classmethod
    def _validate_config(cls, config):
        """
        Helper method for preventing config errors
        """
        if not config.get('DB_TYPE'):
            raise ConfigErr('Please verify that the config specifies'
                            ' the DB_TYPE parameter.')

    @classmethod
    def generate(cls, config, inputdir, outputdir, partner):
        """
        Read the "phi_hashes.csv" file and generate UUID's.

        Optionally save the "hash -> UUID" mapping to "links.csv"
        """
        cls._validate_config(config)
        engine = db.get_db_engine(config)

        # pass a session object to avoid creating it in the loop
        session = db.get_db_session(engine, create_tables=True)

        EXPECTED_COLS = config['EXPECTED_COLS']
        SAVE_OUT_FILE = config['SAVE_OUT_FILE']
        in_file = os.path.join(inputdir, config['IN_FILE'])

        reader = None

        try:
            reader = pd.read_csv(in_file,
                                 sep=config['IN_DELIMITER'],
                                 dtype=object,
                                 skipinitialspace=True,
                                 skip_blank_lines=True,
                                 usecols=list(EXPECTED_COLS),
                                 chunksize=config['LINES_PER_CHUNK'],
                                 iterator=True)
            cls.log.info("Reading data from file: {} ({})"
                         .format(in_file, utils.get_file_size(in_file)))

        except ValueError as exc:
            cls.log.info("Please check if the actual column names"
                         " in [{}] match the expected column names"
                         " file: {}.".format(in_file,
                                             sorted(EXPECTED_COLS)))
            cls.log.error("Error: {}".format(exc))

        frames = []

        for df_source in reader:
            df_source.fillna('', inplace=True)
            # The magic happens here...
            df = cls._process_frame(config, session, df_source, partner)

            if SAVE_OUT_FILE:
                frames.append(df)

        if SAVE_OUT_FILE:
            df = pd.concat(frames, ignore_index=True)
            out_file = os.path.join(outputdir, config['OUT_FILE'])
            utils.frame_to_file(df, out_file,
                                delimiter=config['OUT_DELIMITER'])
            cls.log.info("Wrote output file: {} ({} data rows, {})"
                         .format(
                             out_file, len(df), utils.get_file_size(out_file)))
        return True



"""
Goal: Implement convenience class for managing configurations.

Copied from https://github.com/pallets/flask/blob/master/flask/config.py#L110
"""
import os
import types
import errno
import importlib


class Config(dict):
    """"
    :arg root_path: the prefix for building
        paths to the config files when calling :meth:from_pyfile
    """
    def __init__(self, root_path, defaults=None):
        dict.__init__(self, defaults or {})
        self.root_path = root_path

    def from_pyfile(self, filename, silent=False):
        """Updates the values in the config from a Python file.

        :param filename: the filename of the config.  This can either be an
                         absolute filename or a filename relative to the
                         root path.
        :param silent: set to ``True`` if you want silent failure for missing
                       files.

        """
        filename = os.path.join(self.root_path, filename)
        d = types.ModuleType('config')
        d.__file__ = filename
        try:
            with open(filename) as config_file:
                exec(compile(config_file.read(), filename, 'exec'), d.__dict__)
        except IOError as e:
            if silent and e.errno in (errno.ENOENT, errno.EISDIR):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        self.from_object(d)
        return True

    def from_object(self, obj):
        """Updates the values from the given object.  An object can be of one
        of the following two types:
        -   a string: in this case the object with that name will be imported
        -   an actual object reference: that object is used directly
        :param obj: an import name or object
        """
        string_types = (str,)

        if isinstance(obj, string_types):
            # github.com/pallets/werkzeug/blob/master/werkzeug/utils.py+L399
            # obj = import_string(obj)
            obj = importlib.import_module(obj)

        for key in dir(obj):
            if key.isupper():
                self[key] = getattr(obj, key)

    def from_dictionary(self, dictionary):
        """
        Helper function for overriding config values
        """
        for key, value in dictionary.items():
            if key.isupper():
                self[key] = value



"""
ORM for "rule" table

@authors:
    Andrei Sura <sura.andrei@gmail.com>

+------------------+------------------+------+-----+---------+
| Field            | Type             | Null | Key | Default |
+------------------+------------------+------+-----+---------+
| rule_id          | int(10) unsigned | NO   | PRI | NULL    |
| rule_code        | varchar(255)     | NO   | UNI | NULL    |
| rule_description | varchar(255)     | NO   |     | NULL    |
| rule_added_at    | datetime         | NO   |     | NULL    |
+------------------+------------------+------+-----+---------+
"""

import sqlalchemy as db
from onefl.models.base import DeclarativeBase
from onefl.models.crud_mixin import CRUDMixin

__all__ = ['RuleEntity']


class RuleEntity(CRUDMixin, DeclarativeBase):

    """ Store rules used by the partners sending us the data """
    __tablename__ = 'RULE'

    id = db.Column('RULE_ID', db.Integer, primary_key=True)
    rule_code = db.Column('RULE_CODE', db.Text,
                          nullable=False, unique=True)
    rule_description = db.Column('rule_description', db.Text,
                                 nullable=False)
    rule_added_at = db.Column('RULE_ADDED_AT', db.DateTime,
                              nullable=False)

    @staticmethod
    def get_rules_cache(session):
        """
        :return dictionary: with all available {rule_code -> rule.id} mappings
        """
        rules = session.query(RuleEntity).all()
        result = {rule.rule_code: rule.id for rule in rules}
        return result

    def __repr__(self):
        """ Return a friendly object representation """
        return "<RuleEntity(rule_id: {0.id}, "\
            "rule_code: {0.rule_code}, " \
            "rule_description: {0.rule_description}, " \
            "rule_added_at: {0.rule_added_at})>".format(self)



"""
Goal: expose one method for getting a logger
"""
import os
import logging
import logging.config
from onefl.version import __version__


def get_figlet(name):

    if 'hasher.py' in name:
        return """
 _   _           _
| | | | __ _ ___| |__   ___ _ __
| |_| |/ _` / __| '_ \ / _ \ '__|
|  _  | (_| \__ \ | | |  __/ |
|_| |_|\__,_|___/_| |_|\___|_|   {}
"""

    if 'linker.py' in name:
        return """
_     _       _
| |   (_)_ __ | | _____ _ __
| |   | | '_ \| |/ / _ \ '__|
| |___| | | | |   <  __/ |
|_____|_|_| |_|_|\_\___|_|   {}
"""
    return ''


def get_a_logger(name):
    """
    Read log config from `log.conf` if file is present.
    See the `log.conf.example` for a sample.

    :param name: string used for naming the logger
    """
    basedir, f = os.path.split(__file__)
    log_conf = os.path.join(os.path.dirname(basedir), "config/logs.cfg")

    if os.path.exists(log_conf):
        logging.config.fileConfig(log_conf)
        log = logging.getLogger(name)
        figlet = get_figlet(name)

        if figlet:
            print(figlet.format(__version__))

        log.debug("{} logging was configured using: {}"
                  .format(name, log_conf))
        handlers = logging.getLoggerClass().root.handlers

        for h in [h for h in handlers if hasattr(h, 'baseFilename')]:
            log.debug("log file: {}".format(h.baseFilename))
    else:
        lformat = '%(asctime)s: %(name)s.%(levelname)s ' \
            '- %(filename)s+%(lineno)d: %(message)s'
        logging.basicConfig(format=lformat, level=logging.INFO)
        log = logging.getLogger(name)
        log.warn("logging was configured using defaults")

    return log



"""
ORM for "partner" table

@authors:
    Andrei Sura <sura.andrei@gmail.com>
"""

import sqlalchemy as db
from onefl.models.base import DeclarativeBase
from onefl.models.crud_mixin import CRUDMixin


"""
+---------------------+------------------+------+-----+---------+
| Field               | Type             | Null | Key | Default |
+---------------------+------------------+------+-----+---------+
| partner_code        | varchar(3)       | NO   | UNI | NULL    |
| partner_description | varchar(255)     | NO   |     | NULL    |
| partner_added_at    | datetime         | NO   |     | NULL    |
+---------------------+------------------+------+-----+---------+
"""

__all__ = ['PartnerEntity']


class PartnerEntity(CRUDMixin, DeclarativeBase):

    """ Store partners sendig us the data """
    __tablename__ = 'PARTNER'

    partner_code = db.Column('PARTNER_CODE', db.Text(3), primary_key=True)
    partner_description = db.Column('PARTNER_DESCRIPTION', db.Text,
                                    nullable=False)
    partner_added_at = db.Column('PARTNER_ADDED_AT', db.DateTime,
                                 nullable=False)

    def __repr__(self):
        """ Return a friendly object representation """
        return "<PartnerEntity (partner_code: {0.partner_code}, " \
            "partner_description: {0.partner_description}, " \
            "partner_addded_at: {0.partner_added_at})>".format(self)



"""
Goal: Implement a helper class used for cleaning patient data
    elements before hashing (to increase the chance of a match)

Authors:
     Andrei Sura <sura.andrei@gmail.com>
"""
from onefl import utils


class NormalizedPatient():
    """
    Helper class used to normalize the strings by
    removing punctuation and transforming to lowercase.

    .. seealso::

        Called by :meth:`utils.prepare_for_hashing`
    """

    log = None

    @classmethod
    def configure_logger(cls, logger):
        cls.log = logger

    def __init__(self,  *args, **kwargs):
        self.patid = kwargs.get('patid')

        self.pat_first_name = utils.prepare_for_hashing(
            kwargs.get('pat_first_name')
        )
        self.pat_last_name = utils.prepare_for_hashing(
            kwargs.get('pat_last_name')
        )
        self.pat_birth_date = utils.format_date_as_string(
            kwargs.get('pat_birth_date'), utils.FORMAT_DATABASE_DATE
        )
        self.pat_sex = utils.prepare_for_hashing(
            kwargs.get('pat_sex')
        )
        self.pat_race = utils.prepare_for_hashing(
            kwargs.get('pat_race')
        )

    def has_all_data(self, required_attributes):
        """
        :return bool: `True` if all `required_attributes` have a
        non-empty value
        """
        # TODO: should we check for minimum length of each string?
        for attr in required_attributes:
            if not getattr(self, attr, None):
                NormalizedPatient.log.warning(
                    "--> Missing value for attr [{}]".format(attr))
                return False
        return True

    def __repr__(self):
        return "NormalizedPatient " \
            "<patid: {0.patid}, " \
            "pat_first_name: {0.pat_first_name}, " \
            "pat_last_name: {0.pat_last_name}, " \
            "pat_birth_date: {0.pat_birth_date}, " \
            "pat_sex: {0.pat_sex}, " \
            "pat_race: {0.pat_race}>".format(self)



"""
OneFL Deduper
------------

OneFlorda Entity Resolution Software (aka Deduper)
<https://github.com/ufbmi/onefl-deduper> is designed to generate hashes of the
specific patient data elements.

Multiple generated hashes are then matched against the registry of "known"
hashes and associated with a Unique Universal Identifier (aka OneFlorida UUID).

The UUID is used to de-duplicate medical records such as
demographics, procedures, diagnoses, vitals, lab results submitted
from different hospital networks.

"""
import unittest
from pip.req import parse_requirements
from setuptools import setup, find_packages


def get_test_suite():
    """
    Prepare a test-suite callable with:
        python setup.py test
    """
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('tests', pattern='test_*.py')
    return test_suite


def load_requirements():
    """ Helps to avoid storing requirements in more than one file"""
    reqs = parse_requirements('requirements-to-freeze.txt', session=False)
    reqs_list = [str(ir.req) for ir in reqs]
    return reqs_list


# Configure the package build and distribution
#   @see https://github.com/pypa/setuptools_scm
#   @see http://stackoverflow.com/questions/14399534/how-can-i-reference-requirements-txt-for-the-install-requires-kwarg-in-setuptool  # NOQA
#
# To record the files created use:
#   python setup.py install --record files.txt

setup(
    name="deduper",

    use_scm_version=True,
    url="https://github.com/ufbmi/onefl-deduper",
    license="MIT",
    author="Andrei Sura",
    author_email="bmi-developers@ad.ufl.edu",
    description="OneFlorida De-duplication Software",
    long_description=__doc__,
    keywords=["entity resolution", "deduplication", "record linkage"],

    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    platforms="any",  # darwin, linux2
    setup_requires=["setuptools_scm"],
    install_requires=load_requirements(),
    # tests_require=["mock", "pytest-cov"],  # requirements included above
    test_suite="setup.get_test_suite",

    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
    ],

    entry_points={
        "console_scripts": [
            "hasher = run.hasher:main",
            "linker = run.linker:main",
        ],
    },
)



"""
Goal: implement tests for database models

@authors:
    Andrei Sura <sura.andrei@gmail.com>
"""
import unittest
import binascii
from datetime import datetime
from sqlalchemy.orm.exc import MultipleResultsFound  # noqa
from sqlalchemy.orm.exc import NoResultFound  # noqa

from base_test import BaseTestCase
from onefl import utils
from onefl.models.partner_entity import PartnerEntity  # noqa
from onefl.models.linkage_entity import LinkageEntity
from onefl.models.rule_entity import RuleEntity
from onefl.rules import RULE_CODE_F_L_D_R, RULE_CODE_F_L_D_S  # noqa


class TestModels(BaseTestCase):

    def setUp(self):
        super(TestModels, self).setUp()

    def create_links(self):
        """ Verify we can store LINKAGE rows """
        pers_uuid = utils.get_uuid_bin()

        # For the real code we cun just copy the
        ahash = binascii.unhexlify('2B2D67AED8D511E6A41AF45C898E9B67'.encode())
        added_date = datetime.now()
        cache = RuleEntity.get_rules_cache(self.session)
        rule_id = cache.get(RULE_CODE_F_L_D_R)

        link = LinkageEntity.create(
            partner_code='UFH',
            rule_id=rule_id,
            linkage_patid='123',
            linkage_flag=0,
            linkage_uuid=pers_uuid,
            linkage_hash=ahash,
            linkage_added_at=added_date)
        self.assertIsNotNone(link.id)

        print(link)

        # Search links matching a hash -- should return at most one row
        links_by_hash = self.session.query(LinkageEntity).filter_by(
            linkage_hash=ahash).all()
        self.assertIsNotNone(links_by_hash)

    def test_it(self):
        self.create_links()


if __name__ == '__main__':
    unittest.main()
